/*
 * differential timing analysis simulation.
 *
 * Copyright (c) 2003 by bennet yee <bsy@cs.ucsd.edu>, all rights reserved.
 * see dta.h for more information.
 *
 */
#include <stdio.h>
#include <math.h>
#include <float.h>
#include <stdlib.h>
#include <assert.h>

#include "dta.h"

#define	RANGE_SAMPLES	1024

/*
 * exponentially decayed average.
 */
typedef struct decay {
	double		avg, decay_const;
} decay_t;

inline static void	decay_init(decay_t	*dp,
				   double	decay)
{
	dp->avg = FLT_MIN;
	dp->decay_const = decay;
}

inline static double	decay_value(decay_t	*dp)
{
	return dp->avg;
}

inline void static	decay_new_val(decay_t	*dp,
				      double	v)
{
	if (dp->avg == FLT_MIN) {
		dp->avg = v;
	} else {
		dp->avg = dp->avg * dp->decay_const
			+ v * (1.0 - dp->decay_const);
	}
}

/*
 * "normal" average, standard deviation, etc.
 */
typedef struct statistics {
	double		sum, sumsq;
	uint32_t	nsamp;
} stats_t;

inline static void	stats_init(stats_t	*stp)
{
	stp->sum = stp->sumsq = 0;
	stp->nsamp = 0;
}

inline static double	stats_mean(stats_t	*stp)
{
	return stp->sum / (double) stp->nsamp;
}

inline static void	stats_add_sample(stats_t	*stp,
					 sim_time_t	val)
{
	stp->sum += val;
	stp->sumsq += ((double) val) * val;
	stp->nsamp++;
}

/* sample variance */
inline static double	stats_variance(stats_t	*stp)
{
	return (stp->nsamp * stp->sumsq - stp->sum * stp->sum)/
		((double) stp->nsamp * (stp->nsamp-1));
}

/* sample standard deviation */
inline static double	stats_stdev(stats_t	*stp)
{
	return sqrt(stats_variance(stp));
}

inline static double	stats_std_err_mean(stats_t	*stp)
{
	return stats_stdev(stp)/sqrt(stp->nsamp);
}

exp_t		guessed_exp;
stats_t		bucket[2];
sim_time_t	cutoff, range;

void	print_stats_one(char	*name,
			stats_t	*stp)
{
	printf("%s mean: %17.13e, %d samples, SEM %17.13e\n",name,stats_mean(stp),stp->nsamp,stats_std_err_mean(stp));
}

void	print_stats(void)
{
	print_stats_one("hi",&bucket[1]);
	print_stats_one("lo",&bucket[0]);
}

void	dta(exp_t	*out)
{
	sim_time_t	lo, hi;	/* find max and min to determine cutoff */
	sim_time_t	t;
	int		i;
	big_int_t	prod,x,y;

	lo = SIM_TIME_MAX; hi = SIM_TIME_MIN;

	for (i = 0; i < RANGE_SAMPLES; i++) {
		bigrand(&x);
		bigrand(&y);
		modmult(&prod,&t,&x,&y,modulus);
		if (t < lo)
			lo = t;
		if (hi < t)
			hi = t;
	}
	range = hi - lo;
	cutoff = lo + range/2; /* = avg */

	if (verbose) {
		printf("min mod mult time %d\n",lo);
		printf("max mod mult time %d\n",hi);
		printf("cutoff set to %d\n",cutoff);
		printf("range set to  %d\n",range);
	}

	/* START code for the attack */

	/* END code for the attack */

	*out = guessed_exp;
}
